// hueControl digitalSTROM server add-on
// for controlling Philips Hue LED lamps
// (c) 2012 by plan44.ch
//
// Shared stuff for both background and config


// create two dummy loggers that do nothing
var lWarn = {logln:function (o){}}; 
var lInfo = {logln:function (o){}};

// the name of the logfile
var LOGFILE_NAME = "huecontrol-config.log";
// of the following, comment those out for which we don't want logging
lWarn = new Logger(LOGFILE_NAME);
//lInfo = new Logger(LOGFILE_NAME);


// cURL GET to given URL
// @param aURL full URL
// @return result of query     
function curlGET(aURL)
{ 
  lInfo.logln("curlGET from " + aURL);
  var c = new easycurl;
  c.setopt(c.CURLOPT_URL, aURL);
  c.setopt(c.CURLOPT_TIMEOUT, 4);
  c.setopt(c.CURLOPT_CONNECTTIMEOUT, 3);
  try {
    var resultText = c.perform();  
    lInfo.logln("- returns: " + resultText);
    return resultText;
  }
  catch (err) {
    lWarn.logln("- cURL error: "+err.message);
    return false;
  }
}


// cURL PUT or POST data to given URL
// @param aURL full URL
// @param aPUTdata string to be sent in PUT
// @return result of query     
function curlSEND(aURL, aSENDdata, aMethod)
{
  lInfo.logln("curlSEND to " + aURL + ": " + aSENDdata);
  if (typeof(aMethod)==='undefined') aMethod='PUT';
  var c = new easycurl;
  c.setopt(c.CURLOPT_URL, aURL);
  c.setopt(c.CURLOPT_TIMEOUT, 4);
  c.setopt(c.CURLOPT_CONNECTTIMEOUT, 3);
  c.setopt(c.CURLOPT_POSTFIELDSIZE, aSENDdata.length);
  c.setopt(c.CURLOPT_COPYPOSTFIELDS, aSENDdata);
  if (aMethod=='POST') {
    c.setopt(c.CURLOPT_POST, 1); // use POST 
  }
  else {
    c.setopt(c.CURLOPT_CUSTOMREQUEST, aMethod); // use custom method
  }
  try {
    var resultText = c.perform();  
    lInfo.logln("- returns: " + resultText);
    return resultText;
  }
  catch (err) {
    lWarn.logln("- cURL error: "+err.message);
    return false;
  }
}


// get base URL to access hue bridge
// @return false if no URL found
function hueBridgeBaseURL()
{
  var baseURL = Property.getProperty('bridgeURL');
  if (!baseURL || baseURL.length==0) {
    lWarn.logln("No bridge URL configured");
    return false;
  }
  return baseURL + '/api';
}


// get URL to access hue API
// @return false if bridge cannot be accessed (no URL or no api token;
//         however does not verify API token if one is in the config. Use
//         hueCheckResult() after accessing the API to make sure
//         invalid token is removed from config)
function hueBridgeAccessURL()
{
  var baseURL = hueBridgeBaseURL();
  if (!baseURL) {
    return false;
  }
  // check if there is a api token
  var apiToken = Property.getProperty('bridgeApiToken');
  if (!apiToken || apiToken.length==0) {
    // need to establish connection with the bridge
    lInfo.logln("No API token, must perform login first");
    return false;
  }
  return baseURL + '/' + apiToken;
}


// check hue result and clear API token if access fails due to invalid token
// @param aHueResult hue result object (already JSON parsed)
// @return true on success, false on error 
function hueCheckResult(aHueResult)
{
  lInfo.logln("hueCheckResult: " + JSON.stringify(aHueResult));
  // status result is an array of result objects. Also tolerate query results in other formats
  for (var ei=0; ei<aHueResult.length; ei++) {
    var res = aHueResult[ei];
    if (res) {
      var err = res.error;
      if (err) {
        // error
        if (err.type==1 && err.description=='unauthorized user') {
          lInfo.logln("- login error -> need to obtain new login token");
          // clear API token only on explicit hue bridge login failure!
          Property.getProperty('bridgeApiToken');
          return false;
        }  
      }      
    }
  } // for all returned objects
  return true;
}


// check hue result (text format input variant of hueCheckResult)
// @param aHueResultText hue result text (will be JSON parsed)
// @return true on success, false on error 
function hueCheckResultText(aHueResultText)
{
  if (aHueResultText && aHueResultText.length>0) {
    var hueResult = JSON.parse(aHueResultText);
    return hueCheckResult(hueResult);
  }
  // invalid response
  return false;
}


// issue a GET query to the hue bridge
// @param aQueryAddress - the path and query part of the URL (excluding base API URL and API token)
// @return query result object or false in case of error
function hueQuery(aQueryAddress)
{
  var url = hueBridgeAccessURL();
  var result = false;
  if (url!==false) {
    var resultText = curlGET(url + aQueryAddress);
    if (resultText) {
      result = JSON.parse(resultText);
      if (result) {
        // got JSON result - check for basic error
        if (!hueCheckResult(result))
          return false; // error
        // otherwise, pass on actual result
      }
    }
  }
  return result;
}



// issue a GET query to the hue bridge
// @param aCmdAddress the path and query part of the URL (excluding base API URL and API token)
// @param aCmdJSON command data object to be sent, will be converted to JSON
// @return true on success, false on error
function hueCommand(aCmdAddress, aCmd)
{
  lInfo.logln("hueCommand " + aCmdAddress + " (" + JSON.stringify(aCmd) + ")");
  var url = hueBridgeAccessURL() + aCmdAddress;
  if (url!==false) {
    var data = JSON.stringify(aCmd); // make JSON text from cmd object
    var resultText = curlSEND(url,data);
    if (resultText) {
      // return true or false depending on returned status
      return hueCheckResultText(resultText);
    }
  }
  // if we get here, command has failed anyway
  return false;
}





// register with hue bridge (user needs to press button)
// @return false if failed
function hueBridgeRegisterApp()
{
  var apitoken = false;
  var bridgeApiTokenPath = 'bridgeApiToken';
  // before registering, verify we are not yet successfully registered already
  var cfg = hueQuery('/lights');
  if (cfg) {
    lInfo.logln("hueBridgeRegisterApp: already registered, lights = " + JSON.stringify(cfg));
    return true; // we can query config, so we're registered - no need to re-register
  }
  // can't query config, so we might need to register
  var url = hueBridgeBaseURL();
  lInfo.logln("hueBridgeRegisterApp: base url = " + url);
  if (url) {
    // we have an URL, so try register there
    // - base user name on our mac address, should be unique
    var macaddr = Property.getProperty("/system/host/interfaces/eth0/mac");
    var username = 'dSS';
    for (var i=0; i<macaddr.length; i+=3) {
      username += macaddr.substr(i, 2);
    }
    lInfo.logln("hueBridgeRegisterApp with user name: " + username);
    // first just try with this username, we might already be registered
    var cfg = curlGET(url+username+'/lights');
    if (hueCheckResultText(cfg)) {
      // username is ok as API token
      lInfo.logln("- username is ok as API token, no re-registering needed: " + username);
      apitoken = username;
    }
    else {
      // need real registration
      var regData = {
        "username": username,
        "devicetype": "plan44.ch huecontrol digitalSTROM add-on"
      };
      var regDataText = JSON.stringify(regData);
      var regResultText = curlSEND(url, regDataText, 'POST');
      lInfo.logln("- bridge returns: " + regResultText);
      if (regResultText) {
        var regResult = JSON.parse(regResultText);
        if (regResult) {
          var statusObj = regResult[0];
          if (statusObj) {
            var successObj = statusObj.success;
            if (successObj) {
              apitoken = successObj.username;
              lInfo.logln("- registered successfully, new API token is: " + apitoken);
            }
          }
        }
      }
    }
  }
  // save new API token if we have one
  if (apitoken) {
    Property.setProperty(bridgeApiTokenPath, apitoken);
    Property.setFlag(bridgeApiTokenPath, 'ARCHIVE', true);
    Property.store();
    lInfo.logln("- saved new API token to tree");
    return true; // we are registered now!
  }
  else {
    Property.setProperty(bridgeApiTokenPath, '');    
    Property.setFlag(bridgeApiTokenPath, 'ARCHIVE', true);
    Property.store();
    lWarn.logln("- not registered, no API token");
    return false; // not registered
  }
}






// get list of available lights from bridge
// @return array of light objects with "lightID" and "name"
function hueLights()
{
  var lights = hueQuery('/lights');
  var lightsObjs = [];
  if (!lights) return false; // none
  // create array of light IDs
  lInfo.logln("hueLights: bridge returns: " + JSON.stringify(lights));
  for (var light in lights) {
    lightsObjs.push({ "lightID":light, "name":lights[light].name });
  }
  return lightsObjs;
}




// default parameters for scene
function defaultLampParamsForScene(aSceneNo) {
  switch (aSceneNo) {
    case 0:
    case 10:
    case 20:
    case 30:
    case 40:
    case 72:
       return { "on": false }; // off
    case 1:
    case 11:
    case 21:
    case 31:
    case 41:
    case 71:
       return { "on": true, "ct": 369, "bri":254 }; // standard white (hue poweron default)
    case 2:
    case 12:
    case 22:
    case 32:
    case 42:
       return { "on": true, "ct": 154, "bri":254 }; // coolest white
    case 3:
    case 13:
    case 23:
    case 33:
    case 43:
    case 65:
    case 74:
    case 76:
       return { "on": true, "xy": [ 0.6205, 0.3533 ], "bri":180 }; // red, dimmed
    case 4:
    case 14:
    case 24:
    case 34:
    case 44:
    case 73:
       return { "on": true, "xy": [ 0.1855, 0.0765 ], "bri":180 }; // blue, dimmed
    default:
       return { "on": true, "ct": 154, "bri":254 }; // coolest white
  }
}

function defaultDontCareForScene(aSceneNo) {
  switch (aSceneNo) {
    case 65:
    case 71:
    case 73:
    case 74:
    case 76:
       return 'true';
    default:
       return 'false';
  }
}

// get (and create if not existing) config property for given light ID
function lightConfigProperty(aLightID,aDefaultLightName)
{
  var lightPath = 'lights/light'+String(aLightID);
  var lightConfig = Property.getNode(lightPath);
  if (!lightConfig) {
    // there's no such light config (yet) - create default for it
    lInfo.logln('no config found for light ' + String(aLightID) + '; creating config now');
    lightConfig = new Property(lightPath);
    lightConfig.setFlag('ARCHIVE', true);
    // set ID
    var lightIDPath = lightPath + '/lightID';
    Property.setProperty(lightIDPath,aLightID);
    Property.setFlag(lightIDPath, 'ARCHIVE', true);
    // set name
    var lightNamePath = lightPath + '/name';
    if (typeof(aDefaultLightName!=='undefined') && aDefaultLightName.length>0)
      Property.setProperty(lightNamePath, aDefaultLightName);
    else
      Property.setProperty(lightNamePath, 'light' + String(aLightID));
    Property.setFlag(lightNamePath, 'ARCHIVE', true);
    // put into a zone
    var zoneIDPath = lightPath + '/zoneID';      
    Property.setProperty(zoneIDPath, 0); // default to appartment
    Property.setFlag(zoneIDPath, 'ARCHIVE', true);
    // create a default setting for each scene
    for (var ni=0; ni<=4; ni++) {
      for (var nsi=0; nsi<=4; nsi++) {
        var si = (ni*10)+nsi;
        var scenePath = lightPath + '/scene' + String(si);
        var ns = new Property(scenePath);
        lInfo.logln('- creating scene ' + String(si) + ' for light ' + String(aLightID));
        ns.setFlag('ARCHIVE', true);
        // set ID
        var sceneIDPath = scenePath + '/sceneID';
        Property.setProperty(sceneIDPath, si);
        Property.setFlag(sceneIDPath, 'ARCHIVE', true);
        // set Don't Care
        var sceneDontCarePath = scenePath + '/dontCare';
        Property.setProperty(sceneDontCarePath, defaultDontCareForScene(si));
        Property.setFlag(sceneDontCarePath, 'ARCHIVE', true);
        // set name
        var sceneNamePath = scenePath + '/name';
        Property.setProperty(sceneNamePath, si==0 ? 'Off' : 'Scene' + String(si)); // scene name (Off, Scene1..4)
        Property.setFlag(sceneNamePath, 'ARCHIVE', true);
        // set default params for scene
        var defScene = defaultLampParamsForScene(si);
        var sceneStatePath = scenePath + '/state';
        lInfo.logln('- scene data: ' + JSON.stringify(defScene));
        Property.setProperty(sceneStatePath, JSON.stringify(defScene)); // hue state for this scene
        Property.setFlag(sceneStatePath, 'ARCHIVE', true);
      }
    }
    for (var si=65; si<=76; si++) {
      // das Gleiche wie in der For-Schleife darber nocheinmal fr die Apartmentszenen
        var scenePath = lightPath + '/scene' + String(si);
        var ns = new Property(scenePath);
        lInfo.logln('- creating scene ' + String(si) + ' for light ' + String(aLightID));
        ns.setFlag('ARCHIVE', true);
        // set ID
        var sceneIDPath = scenePath + '/sceneID';
        Property.setProperty(sceneIDPath, si);
        Property.setFlag(sceneIDPath, 'ARCHIVE', true);
        // set Don't Care
        var sceneDontCarePath = scenePath + '/dontCare';
        Property.setProperty(sceneDontCarePath, defaultDontCareForScene(si));
        Property.setFlag(sceneDontCarePath, 'ARCHIVE', true);
        // set name
        var sceneNamePath = scenePath + '/name';
        Property.setProperty(sceneNamePath, si==0 ? 'Off' : 'Scene' + String(si)); // scene name (Off, Scene1..4)
        Property.setFlag(sceneNamePath, 'ARCHIVE', true);
        // set default params for scene
        var defScene = defaultLampParamsForScene(si);
        var sceneStatePath = scenePath + '/state';
        lInfo.logln('- scene data: ' + JSON.stringify(defScene));
        Property.setProperty(sceneStatePath, JSON.stringify(defScene)); // hue state for this scene
        Property.setFlag(sceneStatePath, 'ARCHIVE', true);
    }

  }
  // return the config
  return lightConfig;
}




function applyStateToLight(state,lightNo)
{
  lInfo.logln("light #" + String(lightNo) + " : applying new state: " + JSON.stringify(state));
  return hueCommand('/lights/' + String(lightNo) + '/state', state);
}


function getLightState(lightNo)
{
  lInfo.logln("light #" + String(lightNo) + " : getting current state:");
  var state = false;
  var result = hueQuery('/lights/' + String(lightNo));
  if (result) {
    state = result.state;
    lInfo.logln("- unfiltered state: " + JSON.stringify(state));
    // filter out relevant properties
    // {"on":true,"bri":254,"hue":14950,"sat":142,"xy":[0.4582,0.4101],"ct":365,"effect":"none","colormode":"ct"}
    delete state.reachable;
    delete state.alert;
    delete state.effect;  
    if (state.colormode!='ct') {
      delete state.ct;
    }
    if (state.colormode!='xy') {
      delete state.xy;
    }
    if (state.colormode!='hue') {
      delete state.hue;
      delete state.sat;    
    }
    delete state.colormode;
    lInfo.logln("- filtered state: " + JSON.stringify(state));
  }
  return state;
}


